#include <stdio.h>
#include <string.h>
#include <time.h>
#include <app/Application.h>
#include <app/MessageRunner.h>
#include <app/Roster.h>
#include <stdlib.h>
#include <unistd.h>

#include <interface/Alert.h>
#include <interface/Box.h>
#include <interface/Button.h>
#include <interface/Font.h>
#include <interface/Menu.h>
#include <interface/MenuItem.h>
#include <interface/MenuBar.h>
#include <interface/ScrollBar.h>
#include <interface/Screen.h>
#include <interface/ScrollBar.h>
#include <interface/ScrollView.h>
#include <interface/Input.h>
#include <interface/PopUpMenu.h>

#include <storage/File.h>
#include <storage/Path.h>     
#include <storage/NodeMonitor.h>     
#include <storage/FindDirectory.h>     

#include <support/Beep.h>

#include "ColumnListView.h"
#include "CLVColumn.h"
#include "SplitPane.h"
#include "ShareStrings.h"
#include "ShareUtils.h"

#include "stringtokenizer/StringTokenizer.h"
#include "message/PortableMessage.h"
#include "reflector/StorageReflectConstants.h"
#include "reflector/StorageReflectUtils.h"
#include "netutil/NetworkUtilityFunctions.h"

#include "ShareWindow.h"
#include "ShareNetClient.h"
#include "ShareFileTransfer.h"
#include "ShareColumn.h"
#include "ShareMIMEInfo.h"
#include "RemoteUserItem.h"
#include "RemoteFileItem.h"

#include "alice.h"

static const char * _defaultServers[] = {
   "12.18.240.15",       // Ben's Linux box (email: creator@octavius.org)
   "bebox.corpcomm.net"  // Pete's FreeBSD box
};

#define CHAT_HISTORY_LENGTH 500   // remember the last 500 items you typed
#define BESHARE_SETTINGS_FILE_NAME "beshare_settings"

// Any connection that hasn't transferred data in >= 5 minutes will be cut
#define MORIBUND_TIMEOUT_SECONDS (5*60)

// Window sizing constraints
#define MIN_WIDTH  600
#define MIN_HEIGHT 280
#define MAX_WIDTH  65535
#define MAX_HEIGHT 65535

// Default window position
#define WINDOW_START_X 30
#define WINDOW_START_Y 100          
#define WINDOW_START_W 640
#define WINDOW_START_H 400          

// Size constants for the non-resizable parts of the GUI
#define USER_LIST_WIDTH    125
#define STATUS_VIEW_WIDTH  300
#define UPPER_VIEW_HEIGHT  27
#define CHAT_VIEW_HEIGHT   125
#define USER_ENTRY_WIDTH   150
#define TEXT_ENTRY_HEIGHT  20

#define SPECIAL_COLUMN_CHAR 0x01

#define FILE_NAME_COLUMN_NAME    str(STR_FILE_NAME_KEY)
#define FILE_OWNER_COLUMN_NAME   str(STR_USER_KEY)
#define FILE_SESSION_COLUMN_NAME str(STR_SESSIONID_KEY)
#define FILE_OWNER_BANDWIDTH_NAME str(STR_CONNECTION_KEY)

#define DEFAULT_COLUMN_WIDTH 40.0f

static uint8 DefaultData[256] =
{
        0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
        0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
        0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
        0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
        0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x3F, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
        0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x3F, 0xFE, 0xFE, 0xFE, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
        0xFF, 0xFF, 0xFF, 0x00, 0x3F, 0xFE, 0xFE, 0x62, 0xFE, 0xFE, 0xFE, 0x00, 0x00, 0xFF, 0xFF, 0xFF,
        0xFF, 0xFF, 0x00, 0xFE, 0xFE, 0xFE, 0xB0, 0x62, 0x62, 0x62, 0xFE, 0xFE, 0xFE, 0x00, 0x00, 0xFF,
        0xFF, 0x00, 0xFE, 0xFE, 0xFE, 0xB0, 0xB0, 0x89, 0xB0, 0x62, 0x62, 0x89, 0xFE, 0xFE, 0x00, 0x00,
        0x00, 0xFE, 0xFE, 0xFD, 0xFE, 0xFE, 0xB0, 0xB0, 0xB0, 0x89, 0xB0, 0xFE, 0xFE, 0x00, 0x00, 0xB0,
        0xB0, 0x00, 0x00, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xB0, 0xB0, 0xFD, 0xFD, 0x29, 0x89, 0x00, 0x00,
        0xFF, 0xB0, 0x00, 0x00, 0x00, 0xFD, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0x00, 0x89, 0x00, 0xB0, 0xB0,
        0xFF, 0xFF, 0xB0, 0x00, 0x89, 0x00, 0x00, 0xFA, 0xFA, 0xFA, 0x00, 0x00, 0x00, 0xB0, 0xFF, 0xFF,
        0xFF, 0xFF, 0xFF, 0xB0, 0x00, 0x62, 0x89, 0x00, 0xFA, 0x00, 0xB0, 0xB0, 0xFF, 0xFF, 0xFF, 0xFF,
        0xFF, 0xFF, 0xFF, 0xFF, 0xB0, 0x00, 0x00, 0x00, 0x00, 0xB0, 0xB0, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
        0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xB0, 0x00, 0xB0, 0xFF, 0xB0, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF
};

/* Subclass of BListView that only clears the background area that doesn't contain items.  This reduces flicker during downloads */
class TransferListView : public BListView
{
public:
   TransferListView(BRect rect) : BListView(rect, NULL, B_MULTIPLE_SELECTION_LIST, B_FOLLOW_ALL_SIDES)
   {
      SetViewColor(B_TRANSPARENT_32_BIT);  // we'll draw in the background, thanks
   }

   virtual void Draw(BRect ur)
   {
      BRect backgroundArea = ur;

      int numItems = CountItems();
      if (numItems > 0) backgroundArea.top = ItemFrame(CountItems()-1).bottom+1.0f;

      if (ur.Intersects(backgroundArea))
      {
         SetHighColor(BeBackgroundGrey);
         FillRect(backgroundArea & ur);
      }
      BListView::Draw(ur);
   }
};

/* Subsclass of BTextView that automatically reflows the text when it is resized */
class ReflowingTextView : public BTextView 
{
public:
   ReflowingTextView(BRect frame, const char *name, BRect textRect, uint32 resizeMask, uint32 flags = B_WILL_DRAW | B_PULSE_NEEDED)
      : BTextView(frame, name, textRect, resizeMask, flags)
   {
      // empty
   }

   virtual void FrameResized(float w, float h)
   {
      BTextView::FrameResized(w, h);
      SetTextRect(Frame());
   }
};

static uint32 ShareMIMEInfoPointerHashFunc(ShareMIMEInfo * const & ptr);
uint32 ShareMIMEInfoPointerHashFunc(ShareMIMEInfo * const & ptr)
{
   return (uint32) ptr;
}

ShareWindow :: ShareWindow() :
	BWindow(BRect(WINDOW_START_X,WINDOW_START_Y,WINDOW_START_X+WINDOW_START_W,WINDOW_START_Y+WINDOW_START_H),"BeShare",B_TITLED_WINDOW,0L),
	BMessageFilter(B_ANY_DELIVERY, B_ANY_SOURCE),
	_queryEnabled(false),
	_isConnecting(false),
	_isConnected(false),
	_bytesShown(0LL),
	_netClient(NULL),
	_defaultBitmap(BRect(0,0,15,15),B_COLOR_8_BIT,DefaultData,false,false),
	_users(CStringHashFunc),
	_activeAttribs(PortableStringHashFunc),
	_columns(CStringHashFunc),
	_attribMenuItems(CStringHashFunc),
	_mimeInfos(CStringHashFunc),
	_emptyMimeInfos(ShareMIMEInfoPointerHashFunc),
	_chatHistoryPosition(-1),
	_maxSimultaneousUploadSessions(5),
	_maxSimultaneousUploadSessionsPerUser(2),
	_maxSimultaneousDownloadSessions(5),
	_uploadBandwidth(0),
	_connectionReaper(NULL),
	_language(LANGUAGE_ENGLISH)
{
   const float vMargin = 5.0f;
   const float hMargin = 5.0f;

   _users.SetCompareFunction(CStringCompareFunc);
   _columns.SetCompareFunction(CStringCompareFunc);
   _attribMenuItems.SetCompareFunction(CStringCompareFunc);
   _mimeInfos.SetCompareFunction(CStringCompareFunc);

   BMessenger toMe(this);

   // Attempt to load settings... 
   BMessage settingsMsg;
   BPath sPath;
   if (find_directory(B_USER_SETTINGS_DIRECTORY, &sPath) == B_NO_ERROR)
   {
      sPath.Append(BESHARE_SETTINGS_FILE_NAME);
      BFile settingsFile(sPath.Path(), B_READ_ONLY);
      if (settingsFile.InitCheck() == B_NO_ERROR) (void) settingsMsg.Unflatten(&settingsFile);
   }

   if (settingsMsg.FindInt32("language", &_language) == B_NO_ERROR) SetLanguage(_language);
   
   // Recall saved window position
   BRect pos;
   if (settingsMsg.FindRect("windowpos", &pos) == B_NO_ERROR)
   {
      // check to make sure the window isn't opening off-screen
      BRect screenBounds;
      {
         BScreen s;
         screenBounds = s.Frame();
      }
      if (pos.left > screenBounds.Width()) pos.left = WINDOW_START_X;
      if (pos.top > screenBounds.Height()) pos.top = WINDOW_START_Y;

      MoveTo(pos.left, pos.top);
      ResizeTo(pos.Width(), pos.Height());
   }

   // Recall our "active columns" from the settings message
   BMessage columnsSubMessage;
   if (settingsMsg.FindMessage("columns", &columnsSubMessage) == B_NO_ERROR)
   {
      char * name;
      type_code type;
      int32 count;
      for (int32 i=0; (columnsSubMessage.GetInfo(B_FLOAT_TYPE, i, &name, &type, &count) == B_NO_ERROR); i++)
      {
         float width;
         if (columnsSubMessage.FindFloat(name, &width) == B_NO_ERROR) _activeAttribs.Put(name, width);
      }
   }
   else
   {
      // Put default columns here (size, name, owner, ?)
      _activeAttribs.Put(FILE_NAME_COLUMN_NAME, 330.0f);
      _activeAttribs.Put(FILE_OWNER_COLUMN_NAME, 100.0f);
      _activeAttribs.Put("beshare:File Size", 60.0f);
   }

   // The directory to download files to
   (void)GetAppSubdir("downloads", _downloadsDir, true);

   // Set up our share-file server thread.  This thread is responsible for
   // accepting incoming connections from other BeShare clients and notifying
   // us so that we can start the file transfer to them.
   // Try a few "well-known" ports first, if we can't get any them, dynamically allocate one
   {
      BMessage newSessionMessage(SHAREWINDOW_COMMAND_NEW_PEER_SESSION); 
      _shareServer = new MessageTransceiverThread(toMe);
      bool listening = false;

      for (int i=DEFAULT_LISTEN_PORT; i<=DEFAULT_LISTEN_PORT+LISTEN_PORT_RANGE; i++)
      {
         if (_shareServer->StartServerThread((i<DEFAULT_LISTEN_PORT+LISTEN_PORT_RANGE)?i:0, newSessionMessage) == B_NO_ERROR) 
         {
            listening = true;
            break; 
         }
      }
      if (listening == false)
      {
         delete _shareServer;  
         _shareServer = NULL;
      }
   }

   // set up the net client: this is what talks to the MUSCLE server for us
   (void)GetAppSubdir("shared", _shareDir, true);
   _netClient = new ShareNetClient(_shareDir, _shareServer ? _shareServer->GetPort() : -1);
   AddHandler(_netClient);

   SetSizeLimits(MIN_WIDTH, MAX_WIDTH, MIN_HEIGHT, MAX_HEIGHT);

   _menuBar = new BMenuBar(BRect(), "Menu Bar");

   BMenu * fileMenu = new BMenu(str(STR_FILE));
   fileMenu->AddItem(_connectMenuItem = new BMenuItem(str(STR_CONNECT_TO_SERVER), new BMessage(SHAREWINDOW_COMMAND_RECONNECT_TO_SERVER), shortcut(SHORTCUT_CONNECT)));
   fileMenu->AddItem(_disconnectMenuItem = new BMenuItem(str(STR_DISCONNECT),      new BMessage(SHAREWINDOW_COMMAND_DISCONNECT_FROM_SERVER), shortcut(SHORTCUT_DISCONNECT)));
   fileMenu->AddItem(new BSeparatorItem);

   fileMenu->AddItem(new BMenuItem(str(STR_OPEN_SHARED_FOLDER),  new BMessage(SHAREWINDOW_COMMAND_OPEN_SHARED_FOLDER), shortcut(SHORTCUT_OPEN_SHARED_FOLDER)));
   fileMenu->AddItem(new BMenuItem(str(STR_OPEN_DOWNLOADS_FOLDER), new BMessage(SHAREWINDOW_COMMAND_OPEN_DOWNLOADS_FOLDER), shortcut(SHORTCUT_OPEN_DOWNLOADS_FOLDER)));
   fileMenu->AddItem(new BSeparatorItem);

   _clearChatItem = new BMenuItem(str(STR_CLEAR_CHAT_LOG), new BMessage(SHAREWINDOW_COMMAND_CLEAR_CHAT_LOG), shortcut(SHORTCUT_CLEAR_CHAT_LOG));
   _clearChatItem->SetEnabled(false);

   fileMenu->AddItem(_clearChatItem);
   fileMenu->AddItem(new BMenuItem(str(STR_RESET_LAYOUT), new BMessage(SHAREWINDOW_COMMAND_RESET_LAYOUT), shortcut(SHORTCUT_RESET_LAYOUT)));
   fileMenu->AddItem(new BSeparatorItem);
   fileMenu->AddItem(new BMenuItem(str(STR_ABOUT_BESHARE), new BMessage(SHAREWINDOW_COMMAND_ABOUT), 0));
   fileMenu->AddItem(new BSeparatorItem);
   fileMenu->AddItem(new BMenuItem(str(STR_QUIT), new BMessage(B_QUIT_REQUESTED), shortcut(SHORTCUT_QUIT)));
   _menuBar->AddItem(fileMenu);

   _attribMenu = new BMenu(str(STR_ATTRIBUTES));
   _menuBar->AddItem(_attribMenu);

   BMenu * settingsMenu = new BMenu(str(STR_SETTINGS));
   _menuBar->AddItem(settingsMenu);

   bool fw;
   if ((settingsMsg.FindBool("firewalled", &fw) == B_NO_ERROR)&&(fw)) _netClient->SetFirewalled(true);

   settingsMenu->AddItem(MakeLimitSubmenu(settingsMsg, SHAREWINDOW_COMMAND_SET_UPLOAD_LIMIT, str(STR_MAX_SIMULTANEOUS_UPLOADS), "uploads", _maxSimultaneousUploadSessions));
   settingsMenu->AddItem(MakeLimitSubmenu(settingsMsg, SHAREWINDOW_COMMAND_SET_UPLOAD_PER_USER_LIMIT, str(STR_MAX_SIMULTANEOUS_UPLOADS_PER_USER), "uploadsperuser", _maxSimultaneousUploadSessionsPerUser));
   settingsMenu->AddItem(MakeLimitSubmenu(settingsMsg, SHAREWINDOW_COMMAND_SET_DOWNLOAD_LIMIT, str(STR_MAX_SIMULTANEOUS_DOWNLOADS), "downloads", _maxSimultaneousDownloadSessions));

   BMenu * bMenu = new BMenu(str(STR_UPLOAD_BANDWIDTH));
   bMenu->SetRadioMode(true);
   settingsMenu->AddItem(bMenu);

   int32 bw;
   if (settingsMsg.FindInt32("bandwidth", &bw) == B_NO_ERROR) _uploadBandwidth = bw;
   AddBandwidthOption(bMenu, "300 baud", 75);
   AddBandwidthOption(bMenu, "14.4 kbps", 14400);
   AddBandwidthOption(bMenu, "28.8 kbps", 28800);
   AddBandwidthOption(bMenu, "33.6 kbps", 33600);
   AddBandwidthOption(bMenu, "57.6 kbps", 57600);
   AddBandwidthOption(bMenu, "ISDN-64k",   64000);
   AddBandwidthOption(bMenu, "ISDN-128k", 128000);
   AddBandwidthOption(bMenu, "DSL",       384000);
   AddBandwidthOption(bMenu, "Cable",     768000);
   AddBandwidthOption(bMenu, "T1",  1500000);
   AddBandwidthOption(bMenu, "T3",  4500000);

   BMenu * lMenu = new BMenu(str(STR_LOG));
   settingsMenu->AddItem(lMenu);

   bool lu;

   _logUploads      = new BMenuItem(str(STR_UPLOADS),       new BMessage(SHAREWINDOW_COMMAND_TOGGLE_LOG_UPLOADS));
   if (settingsMsg.FindBool("loguploads", &lu) == B_NO_ERROR) _logUploads->SetMarked(lu);
   lMenu->AddItem(_logUploads);

   _logUserEvents   = new BMenuItem(str(STR_USER_EVENTS),   new BMessage(SHAREWINDOW_COMMAND_TOGGLE_LOG_USER_EVENTS));
   if (settingsMsg.FindBool("loguserevents", &lu) == B_NO_ERROR) _logUserEvents->SetMarked(lu);
                                                            else _logUserEvents->SetMarked(true);  // default to enabled
   lMenu->AddItem(_logUserEvents);

   _logQueryChanges = new BMenuItem(str(STR_QUERY_CHANGES), new BMessage(SHAREWINDOW_COMMAND_TOGGLE_LOG_QUERY_CHANGES));
   if (settingsMsg.FindBool("logquerychanges", &lu) == B_NO_ERROR) _logQueryChanges->SetMarked(lu);
                                                              else _logQueryChanges->SetMarked(true);  // default to enabled
   lMenu->AddItem(_logQueryChanges);

   _logChat = new BMenuItem(str(STR_CHAT), new BMessage(SHAREWINDOW_COMMAND_TOGGLE_LOG_CHAT));
   if (settingsMsg.FindBool("logchat", &lu) == B_NO_ERROR) _logChat->SetMarked(lu);
                                                      else _logChat->SetMarked(true);  // default
   lMenu->AddItem(_logChat);

   _logTimestamps = new BMenuItem(str(STR_TIMESTAMPS), new BMessage(SHAREWINDOW_COMMAND_TOGGLE_LOG_TIMESTAMPS));
   if (settingsMsg.FindBool("logtimestamps", &lu) == B_NO_ERROR) _logTimestamps->SetMarked(lu);
   lMenu->AddItem(_logTimestamps);

   BMenu * langMenu = new BMenu(str(STR_LANGUAGE));
   langMenu->SetRadioMode(true);
   settingsMenu->AddItem(langMenu);

   for (int l=0; l<NUM_LANGUAGES; l++)
   {
      BMessage * lmsg = new BMessage(SHAREWINDOW_COMMAND_SELECT_LANGUAGE);
      lmsg->AddInt32("language", l);
      BMenuItem * nextLMenu = new BMenuItem(GetLanguageName(l), lmsg);
      if (l == _language) nextLMenu->SetMarked(true);
      langMenu->AddItem(nextLMenu);
   }

   _fullUserQueries = new BMenuItem(str(STR_FULL_USER_QUERIES), new BMessage(SHAREWINDOW_COMMAND_TOGGLE_FULL_USER_QUERIES), shortcut(SHORTCUT_FULL_USER_QUERIES));
   if (settingsMsg.FindBool("fulluserqueries", &lu) == B_NO_ERROR) _fullUserQueries->SetMarked(lu);
   settingsMenu->AddItem(_fullUserQueries);

   _beepWhenMentioned = new BMenuItem(str(STR_BEEP_WHEN_MENTIONED), new BMessage(SHAREWINDOW_COMMAND_TOGGLE_BEEP_WHEN_MENTIONED), shortcut(SHORTCUT_BEEP_WHEN_MENTIONED));
   if (settingsMsg.FindBool("beepwhenmentioned", &lu) == B_NO_ERROR) _beepWhenMentioned->SetMarked(lu);
   settingsMenu->AddItem(_beepWhenMentioned);

   _autoClearCompletedDownloads = new BMenuItem(str(STR_AUTOCLEAR_COMPLETED_DOWNLOADS), new BMessage(SHAREWINDOW_COMMAND_TOGGLE_AUTOCLEAR_COMPLETED_DOWNLOADS), 0);
   if (settingsMsg.FindBool("autoclear", &lu) == B_NO_ERROR) _autoClearCompletedDownloads->SetMarked(lu);
   settingsMenu->AddItem(_autoClearCompletedDownloads);

   settingsMenu->AddItem(new BSeparatorItem);
   _firewalled = new BMenuItem(str(STR_IM_FIREWALLED), new BMessage(SHAREWINDOW_COMMAND_TOGGLE_FIREWALLED));

   settingsMenu->AddItem(_firewalled);

   AddChild(_menuBar);                          

   BRect contentBounds = Bounds();
   contentBounds.top = _menuBar->Bounds().Height();

   // Create group/area views (top level stuff)
   BView * contentView = new BView(contentBounds, "ContentView", B_FOLLOW_ALL_SIDES, 0);
   contentView->SetViewColor(BeBackgroundGrey);
   AddChild(contentView);

   {
      // Fill out the upperLevel view
      BRect upperViewBounds(0, 0, contentBounds.Width() - hMargin, UPPER_VIEW_HEIGHT);
      BView * upperView = new BView(upperViewBounds, "UpperView", B_FOLLOW_LEFT_RIGHT | B_FOLLOW_TOP, 0);
      upperView->SetViewColor(BeBackgroundGrey);
      contentView->AddChild(upperView);

      {
         // Fill out the query view
         BRect queryViewBounds(0, 0, upperViewBounds.Width()-STATUS_VIEW_WIDTH, upperViewBounds.Height());
         BView * queryView = new BView(queryViewBounds, NULL, B_FOLLOW_ALL_SIDES, 0);
         queryView->SetViewColor(BeBackgroundGrey);
         upperView->AddChild(queryView);

         const char * q = str(STR_QUERY);
         _queryMenu = new BMenu(q);
         float qw = queryView->StringWidth(q)+36.0f;
         queryView->AddChild(new BMenuField(BRect(hMargin,4,qw,TEXT_ENTRY_HEIGHT), NULL, NULL, _queryMenu));

         float right = queryViewBounds.Width()-hMargin;
         float stringWidth = _queryMenu->StringWidth(str(STR_STOP_QUERY))+10.0f;
         _disableQueryButton = new BButton(BRect(right-stringWidth,3,right,TEXT_ENTRY_HEIGHT), NULL, str(STR_STOP_QUERY), new BMessage(SHAREWINDOW_COMMAND_DISABLE_QUERY), B_FOLLOW_RIGHT | B_FOLLOW_TOP);
         queryView->AddChild(_disableQueryButton);
         right -= (stringWidth + hMargin);

         stringWidth = _queryMenu->StringWidth(str(STR_START_QUERY))+10.0f;
         _enableQueryButton = new BButton(BRect(right-stringWidth,3,right,TEXT_ENTRY_HEIGHT), NULL, str(STR_START_QUERY), new BMessage(SHAREWINDOW_COMMAND_ENABLE_QUERY), B_FOLLOW_RIGHT | B_FOLLOW_TOP);
         queryView->AddChild(_enableQueryButton);
         right -= (stringWidth + hMargin);

         const char * startupQuery;
         if (settingsMsg.FindString("query", &startupQuery) != B_NO_ERROR) startupQuery = "*.mp3";
         _fileNameQueryEntry = new BTextControl(BRect(qw-10.0f,6,right,TEXT_ENTRY_HEIGHT), NULL, NULL, startupQuery, new BMessage(SHAREWINDOW_COMMAND_CHANGE_FILE_NAME_QUERY), B_FOLLOW_ALL_SIDES);
         _fileNameQueryEntry->SetTarget(toMe);
         queryView->AddChild(_fileNameQueryEntry);

         // Restore any additional strings....
         const char * listQuery;
         for (int qh=1; (settingsMsg.FindString("query", qh, &listQuery) == B_NO_ERROR); qh++) 
         {
            BMessage * msg = new BMessage(SHAREWINDOW_COMMAND_CHANGE_FILE_NAME_QUERY);
            msg->AddString("query", listQuery);
            _queryMenu->AddItem(new BMenuItem(listQuery, msg), 0);
         }
      }

      {
         // Fill out the status view
         BRect statusViewBounds(upperViewBounds.Width()-STATUS_VIEW_WIDTH, 0, upperViewBounds.Width(), upperViewBounds.Height());
         BView * statusView = new BView(statusViewBounds, NULL, B_FOLLOW_RIGHT| B_FOLLOW_TOP_BOTTOM, 0);
         statusView->SetViewColor(BeBackgroundGrey);
         upperView->AddChild(statusView);

         float serverMenuWidth = _queryMenu->StringWidth(str(STR_SERVER))+25.0f;
         _serverMenu = new BMenu(str(STR_SERVER));
         statusView->AddChild(new BMenuField(BRect(0,4,serverMenuWidth,statusViewBounds.Height()), NULL, NULL, _serverMenu));

         const char * firstName = NULL;
         const char * sn = NULL;
         for (int i=0; (settingsMsg.FindString("serverlist", i, &sn) == B_NO_ERROR); i++)
         {
            if (firstName == NULL) firstName = sn;
            AddServerItem(sn);
         }
         if (firstName == NULL)
         {
            firstName = _defaultServers[0];
            for (uint32 j=0; j<ARRAYITEMS(_defaultServers); j++) AddServerItem(_defaultServers[j]);
         }

         if (settingsMsg.FindString("server", &sn) == B_NO_ERROR) firstName = sn;
         _serverEntry = new BTextControl(BRect(serverMenuWidth, 6, STATUS_VIEW_WIDTH-(USER_ENTRY_WIDTH+hMargin), statusViewBounds.Height()), NULL, NULL, firstName, new BMessage(SHAREWINDOW_COMMAND_USER_CHANGED_SERVER));
         _serverEntry->SetTarget(toMe);
         _serverEntry->SetDivider(0.0f);
         statusView->AddChild(_serverEntry);   

         const char * un;
         if (settingsMsg.FindString("username", &un) != B_NO_ERROR) un = str(STR_ANONYMOUS);
         _netClient->SetLocalUserName(un);

         _userNameEntry = new BTextControl(BRect(STATUS_VIEW_WIDTH-USER_ENTRY_WIDTH,6,STATUS_VIEW_WIDTH-1,statusViewBounds.Height()), NULL, str(STR_USER_NAME_COLON), un, new BMessage(SHAREWINDOW_COMMAND_USER_CHANGED_NAME));
         _userNameEntry->SetDivider(_userNameEntry->StringWidth(str(STR_USER_NAME_COLON))+4.0f);
         _userNameEntry->SetTarget(toMe);
         
         statusView->AddChild(_userNameEntry);
      }
   }
    
   BRect middleBounds(2, UPPER_VIEW_HEIGHT, contentBounds.Width()-hMargin, contentBounds.Height()-CHAT_VIEW_HEIGHT);

   BRect resultsBounds(0, 0, middleBounds.Width()-(USER_LIST_WIDTH+hMargin), middleBounds.Height());
   BView * resultsView = new BView(resultsBounds, "IOView", B_FOLLOW_ALL_SIDES, 0);
   resultsView->SetViewColor(BeBackgroundGrey);

   CLVContainerView* resultsContainerView;
   _resultsView = new ColumnListView(BRect(hMargin, vMargin, resultsView->Bounds().Width()-(B_V_SCROLL_BAR_WIDTH+2), resultsView->Bounds().Height()-(vMargin+TEXT_ENTRY_HEIGHT+B_H_SCROLL_BAR_HEIGHT+8)),&resultsContainerView, NULL, B_FOLLOW_ALL_SIDES, B_WILL_DRAW|B_FRAME_EVENTS|B_NAVIGABLE,B_MULTIPLE_SELECTION_LIST,false,true,true,true,B_FANCY_BORDER);
   _resultsView->SetSortFunction((CLVCompareFuncPtr) CompareFunc);
   _resultsView->SetTarget(toMe);
   _resultsView->SetSelectionMessage(new BMessage(SHAREWINDOW_COMMAND_RESULT_SELECTION_CHANGED));
   _resultsView->AddFilter(this);
   resultsView->AddChild(resultsContainerView);

   _resultsView->AddColumn(new CLVColumn("", 20.0f, CLV_LOCK_AT_BEGINNING | CLV_NOT_MOVABLE | CLV_NOT_RESIZABLE));

   _requestDownloadsButton = new BButton(BRect(_resultsView->Bounds().left+2, resultsView->Bounds().Height()-(TEXT_ENTRY_HEIGHT+vMargin+1),resultsBounds.right-250,resultsBounds.Height()-(vMargin)), NULL, str(STR_DOWNLOAD_SELECTED_FILES), new BMessage(SHAREWINDOW_COMMAND_BEGIN_DOWNLOADS), B_FOLLOW_LEFT_RIGHT | B_FOLLOW_BOTTOM, B_WILL_DRAW|B_NAVIGABLE|B_FULL_UPDATE_ON_RESIZE);
   resultsView->AddChild(_requestDownloadsButton);

   _clearFinishedDownloadsButton = new BButton(BRect(resultsBounds.right-245, resultsView->Bounds().Height()-(TEXT_ENTRY_HEIGHT+vMargin+1),resultsBounds.right-2,resultsBounds.Height()-(vMargin)), NULL, str(STR_CLEAR_FINISHED_FAILED_TRANSFERS), new BMessage(SHAREWINDOW_COMMAND_CLEAR_FINISHED_DOWNLOADS), B_FOLLOW_RIGHT | B_FOLLOW_BOTTOM);
   resultsView->AddChild(_clearFinishedDownloadsButton);

   BRect transferBounds(resultsBounds.right+hMargin, resultsBounds.top+3, middleBounds.Width()-hMargin, middleBounds.bottom-30);
   BView * transferView = new BView(transferBounds, NULL, B_FOLLOW_RIGHT | B_FOLLOW_TOP_BOTTOM, 0);
   transferView->SetViewColor(BeBackgroundGrey);

   _transferList = new TransferListView(BRect(0, 2, transferBounds.Width()-(2+B_V_SCROLL_BAR_WIDTH), transferBounds.Height()-(5+TEXT_ENTRY_HEIGHT+vMargin)));
   _transferList->SetTarget(toMe);
   _transferList->SetSelectionMessage(new BMessage(SHAREWINDOW_COMMAND_RESULT_SELECTION_CHANGED));
   _transferList->SetInvocationMessage(new BMessage(SHAREWINDOW_COMMAND_LAUNCH_TRANSFER_ITEM));
   _transferList->AddFilter(this);
   transferView->AddChild(new BScrollView(NULL, _transferList, B_FOLLOW_ALL_SIDES, 0L, false, true, B_FANCY_BORDER));

   _cancelTransfersButton = new BButton(BRect(0, _transferList->Frame().bottom+vMargin-1, transferBounds.Width(), transferBounds.Height()-1), NULL, str(STR_REMOVE_SELECTED), new BMessage(SHAREWINDOW_COMMAND_CANCEL_DOWNLOADS), B_FOLLOW_LEFT_RIGHT | B_FOLLOW_BOTTOM);
   transferView->AddChild(_cancelTransfersButton);

   _resultsTransferSplit = new SplitPane(middleBounds, resultsView, transferView, B_FOLLOW_ALL_SIDES);
   _resultsTransferSplit->SetResizeViewOne(true, true);
   _resultsTransferSplit->SetMinSizeOne(BPoint(380.0f, 90.0f));
   _resultsTransferSplit->SetMinSizeTwo(BPoint(100.0f, 50.0f));
   _resultsTransferSplit->SetBarPosition(BPoint(_resultsTransferSplit->Bounds().Width()-transferView->Bounds().Width(), _resultsTransferSplit->GetBarPosition().y));
   _resultsTransferSplit->SetViewColor(BeBackgroundGrey);

   BRect bottomBounds(hMargin, contentBounds.Height()+vMargin-CHAT_VIEW_HEIGHT, contentBounds.right-hMargin, contentBounds.Height()-vMargin);
   BRect chatViewBounds(0, 0, bottomBounds.Width()-(USER_LIST_WIDTH+hMargin), bottomBounds.Height());
   BView * chatView = new BView(chatViewBounds, NULL, B_FOLLOW_ALL_SIDES, 0); 
   chatView->SetViewColor(BeBackgroundGrey);

   BRect chatTextBounds(0, 0, chatViewBounds.Width(), chatViewBounds.Height()-(TEXT_ENTRY_HEIGHT+vMargin));
   BBox * box = new BBox(chatTextBounds, NULL, B_FOLLOW_ALL_SIDES);
   chatView->AddChild(box);
      
   _chatText = new ReflowingTextView(BRect(2,2,chatTextBounds.Width()-(2+B_V_SCROLL_BAR_WIDTH), chatTextBounds.Height()-2), NULL, chatTextBounds, B_FOLLOW_ALL_SIDES);
   _chatText->MakeEditable(false);
   _chatText->SetStylable(true);
   _chatText->AddFilter(this);

   _chatScrollView = new BScrollView(NULL, _chatText, B_FOLLOW_ALL_SIDES, 0L, false, true, B_FANCY_BORDER);
   box->AddChild(_chatScrollView);

   PortableString chat(str(STR_CHAT));
   chat += ':';
   _textEntry = new BTextControl(BRect(0, chatViewBounds.Height()-TEXT_ENTRY_HEIGHT, chatViewBounds.Width(), chatViewBounds.Height()), NULL, chat.Cstr(), NULL, NULL, B_FOLLOW_LEFT_RIGHT | B_FOLLOW_BOTTOM);
   _textEntry->SetDivider(_textEntry->StringWidth(chat.Cstr())+4.0f);
   _textEntry->SetTarget(toMe);
   _textEntry->TextView()->AddFilter(this);
   chatView->AddChild(_textEntry);

   BView * userListView = new BView(BRect(chatViewBounds.right+hMargin, bottomBounds.top, bottomBounds.right, bottomBounds.bottom), NULL, B_FOLLOW_RIGHT | B_FOLLOW_BOTTOM, 0);
   userListView->SetViewColor(BeBackgroundGrey);

   CLVContainerView* userContainerView;
   _usersView = new ColumnListView(BRect(0, 2, userListView->Bounds().Width()-(B_V_SCROLL_BAR_WIDTH+2), userListView->Bounds().Height()-2),&userContainerView,NULL,B_FOLLOW_ALL_SIDES, B_WILL_DRAW|B_FRAME_EVENTS|B_NAVIGABLE,B_MULTIPLE_SELECTION_LIST,false,false,true,true,B_FANCY_BORDER);
   _usersView->AddColumn(new CLVColumn(str(STR_NAME), _usersView->Bounds().Width()*.70f, CLV_SORT_KEYABLE));
   _usersView->AddColumn(new CLVColumn(str(STR_ID), _usersView->Bounds().Width()*.26f, CLV_SORT_KEYABLE));
   _usersView->AddColumn(new CLVColumn(str(STR_FILES), _usersView->Bounds().Width()*.38, CLV_SORT_KEYABLE));
   _usersView->AddColumn(new CLVColumn(str(STR_CONNECTION_KEY)+2, _usersView->Bounds().Width()*.57, CLV_SORT_KEYABLE));
   _usersView->SetSortFunction((CLVCompareFuncPtr) UserCompareFunc);
   _usersView->SetMessage(new BMessage(SHAREWINDOW_COMMAND_SELECT_USER));
   _usersView->SetTarget(toMe);
   _usersView->AddFilter(this);
   userListView->AddChild(userContainerView);

   _chatUsersSplit = new SplitPane(bottomBounds, chatView, userListView, B_FOLLOW_LEFT_RIGHT | B_FOLLOW_BOTTOM);
   _chatUsersSplit->SetResizeViewOne(true, true);
   _chatUsersSplit->SetMinSizeOne(BPoint(200.0f, 50.0f));
   _chatUsersSplit->SetMinSizeTwo(BPoint(50.0f, 50.0f));
   _chatUsersSplit->SetBarPosition(BPoint(_chatUsersSplit->Bounds().Width()-(20+_usersView->Bounds().Width()), _chatUsersSplit->GetBarPosition().y));
   _chatUsersSplit->SetViewColor(BeBackgroundGrey);

   _mainSplit = new SplitPane(BRect(contentBounds.left, contentBounds.top+vMargin, contentBounds.right, contentBounds.bottom-20), _resultsTransferSplit, _chatUsersSplit, B_FOLLOW_ALL_SIDES);
   _mainSplit->SetResizeViewOne(true, true);
   _mainSplit->SetMinSizeOne(BPoint(300.0f, 90.0f));
   _mainSplit->SetMinSizeTwo(BPoint(100.0f, 70.0f));
   _mainSplit->SetBarPosition(BPoint(contentView->Bounds().Height()/2.0f, _mainSplit->GetBarPosition().y));

   ResetLayout();
   RestoreSplitPane(_resultsTransferSplit, settingsMsg, "resultstransfersplit"); 
   RestoreSplitPane(_chatUsersSplit, settingsMsg, "chatuserssplit"); 
   RestoreSplitPane(_mainSplit, settingsMsg, "mainsplit"); 

   contentView->AddChild(_mainSplit);

   UpdateConnectStatus(true);
   UpdateQueryEnabledStatus();

   // Columns that we know will available for any file we can put up right away
   CreateColumn(NULL, FILE_NAME_COLUMN_NAME,     false, '1');
   CreateColumn(NULL, "beshare:File Size",       false, '2');
   CreateColumn(NULL, "beshare:Kind",            false, '3');
   CreateColumn(NULL, FILE_OWNER_COLUMN_NAME,    false, '4');
   CreateColumn(NULL, FILE_OWNER_BANDWIDTH_NAME, false, '5');
   CreateColumn(NULL, FILE_SESSION_COLUMN_NAME,  false, '6');

_aliceRunner = new BMessageRunner(toMe, new BMessage(SHAREWINDOW_COMMAND_CHECK_ALICE), 12*1000000LL);

   // The reaper will have us examine our transfer sessions every so
   // often, and kill any that are active-but-not-transferring-anything.
   // This prevents someone with a broken connection from piling up
   // a lot of useless requests, DOS'ing your client.
   _connectionReaper = new BMessageRunner(toMe, new BMessage(SHAREWINDOW_COMMAND_CHECK_FOR_MORIBUND_CONNECTIONS), 60*1000000LL); // check once per minute

   if (_shareServer)
   {
      char temp[150];
      sprintf(temp, str(STR_BESHARE_IS_LISTENING_ON_PORT_PERCENTI), _shareServer->GetPort());
      LogMessage(LOG_INFORMATION_MESSAGE, temp);
   }
   else LogMessage(LOG_ERROR_MESSAGE, str(STR_COULDNT_START_FILE_SHARING_THREAD));
}

void
ShareWindow ::
RestoreSplitPane(SplitPane * sp, const BMessage & settingsMsg, const char * name) const
{
   BMessage temp;
   if (settingsMsg.FindMessage(name, &temp) == B_NO_ERROR) sp->SetState(&temp);
}

void
ShareWindow ::
AddServerItem(const char * sn)
{
   BMessage * msg = new BMessage(SHAREWINDOW_COMMAND_USER_SELECTED_SERVER);
   msg->AddString("server", sn);
   _serverMenu->AddItem(new BMenuItem(sn, msg));
}
         
void
ShareWindow ::
SaveSplitPane(const SplitPane * sp, BMessage & settingsMsg, const char * name) const
{
   BMessage state;
   sp->GetState(state);
   settingsMsg.AddMessage(name, &state);
}

status_t
ShareWindow :: GetAppSubdir(const char * subDirName, BDirectory & subDir, bool createIfNecessary) const
{
   app_info appInfo;
   be_app->GetAppInfo(&appInfo);
   BEntry appEntry(&appInfo.ref);
   appEntry.GetParent(&appEntry);  // get the directory it's in
   BPath path(&appEntry);
   BPath subPath(&appEntry);
   subPath.Append(subDirName);

   // If the directory is already there, use it
   if (subDir.SetTo(subPath.Path()) == B_NO_ERROR) return B_NO_ERROR;

   // Directory not there?  Shall we create it then?
   if (createIfNecessary)
   {
      BDirectory appDir(path.Path());
      if ((appDir.InitCheck() == B_NO_ERROR)&&(appDir.CreateDirectory(subDirName, &subDir) == B_NO_ERROR)) return B_NO_ERROR;
   }
   return B_ERROR;  // oops, couldn't get it
}

ShareWindow :: ~ShareWindow()
{
   delete _connectionReaper;
delete _aliceRunner;

   // delete MIME infos that aren't part of our menu hierarchy (the ones in the menu will be deleted by the BMenu)
   ShareMIMEInfo * mi;
   PortableHashtableIterator<ShareMIMEInfo *, bool> miter = _emptyMimeInfos.GetIterator();
   while(miter.GetNextKey(mi)) delete mi;

   // Remove filters to avoid crashing due to dangling pointers
   _resultsView->RemoveFilter(this);
   _chatText->RemoveFilter(this);
   _transferList->RemoveFilter(this);
   _usersView->RemoveFilter(this);
   _textEntry->TextView()->RemoveFilter(this);

   // Save settings...
   BMessage settingsMsg;
   SaveSplitPane(_resultsTransferSplit, settingsMsg, "resultstransfersplit"); 
   SaveSplitPane(_chatUsersSplit, settingsMsg, "chatuserssplit"); 
   SaveSplitPane(_mainSplit, settingsMsg, "mainsplit"); 
  
   settingsMsg.AddInt32("language", _language);

   if (_netClient->GetLocalUserName()[0]) settingsMsg.AddString("username", _netClient->GetLocalUserName());
   if (_fileNameQueryEntry->Text()[0]) settingsMsg.AddString("query", _fileNameQueryEntry->Text());

   // Save any additional strings....
   int qmLen = _queryMenu->CountItems();
   for (int qh=0; qh<qmLen; qh++) settingsMsg.AddString("query", _queryMenu->ItemAt(qh)->Label());

   settingsMsg.AddString("server", _serverEntry->Text());
   settingsMsg.AddBool("loguploads", _logUploads->IsMarked());
   settingsMsg.AddBool("loguserevents", _logUserEvents->IsMarked());
   settingsMsg.AddBool("logquerychanges", _logQueryChanges->IsMarked());
   settingsMsg.AddBool("logchat", _logChat->IsMarked());
   settingsMsg.AddBool("logtimestamps", _logTimestamps->IsMarked());
   settingsMsg.AddBool("fulluserqueries", _fullUserQueries->IsMarked());
   settingsMsg.AddBool("beepwhenmentioned", _beepWhenMentioned->IsMarked());
   settingsMsg.AddBool("autoclear", _autoClearCompletedDownloads->IsMarked());
   
   settingsMsg.AddInt32("uploads", _maxSimultaneousUploadSessions);
   settingsMsg.AddInt32("downloads", _maxSimultaneousDownloadSessions);
   settingsMsg.AddInt32("uploadsperuser", _maxSimultaneousUploadSessionsPerUser);
   settingsMsg.AddInt32("bandwidth", _uploadBandwidth);

   settingsMsg.AddBool("firewalled", _netClient->GetFirewalled());

   for (int i=0; i<_serverMenu->CountItems(); i++) settingsMsg.AddString("serverlist", _serverMenu->ItemAt(i)->Label());

   BRect windowpos = Frame();
   settingsMsg.AddRect("windowpos", windowpos);

   BMessage columnsSubMessage;
   PortableHashtableIterator<PortableString, float> colIter = _activeAttribs.GetIterator();
   PortableString * nextString;
   while((nextString = colIter.GetNextKey()) != NULL) 
   {
      float width = *colIter.GetNextValue();  // default width in case we can't find the column itself for some reason
      float colWidth = 0.0f;

      ShareColumn * sc;
      if (_columns.Get(nextString->Cstr(), sc)) colWidth = sc->Width();
      columnsSubMessage.AddFloat(nextString->Cstr(), (colWidth > 0.0f) ? colWidth : width);
   }
   settingsMsg.AddMessage("columns", &columnsSubMessage);

   // Attempt to save settings... 
   BPath sPath;
   if (find_directory(B_USER_SETTINGS_DIRECTORY, &sPath) == B_NO_ERROR)
   {
      sPath.Append(BESHARE_SETTINGS_FILE_NAME);
      BFile settingsFile(sPath.Path(), B_WRITE_ONLY|B_CREATE_FILE|B_ERASE_FILE);
      if (settingsFile.InitCheck() == B_NO_ERROR) (void) settingsMsg.Flatten(&settingsFile);
   }

   ClearUsers();

   delete _netClient;
   delete _shareServer;
}

void
ShareWindow ::
ClearUsers()
{
   ClearResults();             // no users means no files available
   _usersView->MakeEmpty();    // for efficiency
   PortableHashtableIterator<const char *, RemoteUserItem *> iter = _users.GetIterator();
   RemoteUserItem * next;
   while(iter.GetNextValue(next)) delete next;
   _users.Clear();
}

// Hash function that just uses the pointer value as a hash value
static uint32 RemoteUserItemPointerHash(RemoteUserItem * const & ptr);
static uint32 RemoteUserItemPointerHash(RemoteUserItem * const & ptr)
{
   return (uint32)ptr;
}

BMenu *
ShareWindow ::
MakeLimitSubmenu(const BMessage & settingsMsg, uint32 code, const char * label, const char * fieldName, uint32 & var)
{
   BMenu * qMenu = new BMenu(label);
   qMenu->SetRadioMode(true);

   int32 limit = var;
   if (settingsMsg.FindInt32(fieldName, &limit) == B_NO_ERROR) var = limit;

   uint32 limits[] = {1, 2, 3, 4, 5, 10, 20, 999999};
   for (uint32 i=0; i<ARRAYITEMS(limits); i++)
   {
      char temp[80];
      if (i < ARRAYITEMS(limits)-1) sprintf(temp, "%lu", limits[i]);
                               else strncpy(temp, str(STR_NO_LIMIT), sizeof(temp));
      BMessage * msg = new BMessage(code);
      msg->AddInt32("num", limits[i]);
      BMenuItem * item = new BMenuItem(temp, msg);
      if (limits[i] == (uint32) limit) item->SetMarked(true);
      qMenu->AddItem(item);
   }
   return qMenu;
}

void
ShareWindow ::
AddBandwidthOption(BMenu * bMenu, const char * label, int bps)
{
   BMessage * msg = new BMessage(SHAREWINDOW_COMMAND_SET_ADVERTISED_BANDWIDTH);
   msg->AddInt32("bps", bps);
   msg->AddString("label", label);
   BMenuItem * item = new BMenuItem(label, msg);
   if (bps == (int) _uploadBandwidth) 
   {
      item->SetMarked(true);
      _netClient->SetUploadBandwidth(label, bps);
   }
   bMenu->AddItem(item);
}


void
ShareWindow ::
ConnectBackRequestReceived(const char * targetSessionID, int port)
{
   RemoteUserItem * target;
   if ((_users.Get(targetSessionID, target))&&(target->GetFirewalled() == false))
   {
      ShareFileTransfer * xfer = new ShareFileTransfer(_shareDir, _netClient->GetLocalSessionID());
      AddHandler(xfer);
      
      if (xfer->InitConnectSession(target->GetHostName(), port, target->GetSessionID()) == B_NO_ERROR) 
      {
         _transferList->AddItem(xfer);
      }
      else
      {
         RemoveHandler(xfer);
         delete xfer;
      }
      DequeueTransferSessions();
   }
}

void 
ShareWindow ::
RequestDownloads()
{
   // First, collate the files by remote user.  We do this so that we only have to
   // make one TCP connection to each remote host, and download all files that are
   // to come from him one at a time over that.  It's more efficient than connecting
   // to him several times in parallel.
   PortableHashtable<RemoteUserItem *, ShareFileTransfer *> newTransferSessions(RemoteUserItemPointerHash);
   int32 nextIndex;
   for (int32 i=0; ((nextIndex =_resultsView->CurrentSelection(i)) >= 0); i++)
   {
      RemoteFileItem * item = (RemoteFileItem *) _resultsView->ItemAt(nextIndex);
      RemoteUserItem * owner = item->GetOwner();
      if ((owner->GetHostName()[0])&&(owner->GetPort() > 0))
      {
         ShareFileTransfer * xfer;
         if (newTransferSessions.Get(owner, xfer) == false)
         {
            xfer = new ShareFileTransfer(_downloadsDir, _netClient->GetLocalSessionID());
            AddHandler(xfer);
            newTransferSessions.Put(owner, xfer);
         }
         xfer->AddRequestedFileName(item->GetFileName(), 0LL);
      }
      else 
      {
         PortableString errStr(str(STR_CANT_DOWNLOAD_FROM_USER));
         errStr += owner->GetHandle();
         errStr += str(STR_COMMA_NO_CONNECTION_INFORMATION_AVAILABLE);
         LogMessage(LOG_ERROR_MESSAGE, errStr.Cstr());
      }
   }

   // Set up the ShareFileTransfers to await their incoming connections from the remote owners...
   PortableHashtableIterator<RemoteUserItem *, ShareFileTransfer *> iter = newTransferSessions.GetIterator();
   ShareFileTransfer * nextXfer;
   while(iter.GetNextValue(nextXfer))
   {
      bool okayToAdd = false;

      RemoteUserItem * nextUser = *(iter.GetNextKey());
      if (nextUser->GetFirewalled())
      {
         if (_netClient->GetFirewalled())
         {
            PortableString errStr(str(STR_CANT_DOWNLOAD_FILES_FROM));
            errStr += nextUser->GetUserString();
            errStr += str(STR_BECAUSE_BOTH_OF_US_ARE_BEHIND_FIREWALLS);
            LogMessage(LOG_ERROR_MESSAGE, errStr.Cstr());
         }
         else
         {
            if (nextXfer->InitAcceptSession(nextUser->GetSessionID()) == B_NO_ERROR) okayToAdd = true;
            else
            {
               PortableString errStr(str(STR_FILE_DOWNLOAD_ACCEPT_SESSION_FOR));
               errStr += nextUser->GetUserString();
               errStr += str(STR_FAILED_TO_INITIALIZE);
               LogMessage(LOG_ERROR_MESSAGE, errStr.Cstr());
            }
         }
      }
      else
      {
         // He's not firewalled so we can connect to him directly.
         if (nextXfer->InitConnectSession(nextUser->GetHostName(), nextUser->GetPort(), nextUser->GetSessionID()) == B_NO_ERROR) okayToAdd = true;
         else
         {
            PortableString errStr(str(STR_FILE_DOWNLOAD_SESSION_TO));
            errStr += nextUser->GetUserString();
            errStr += str(STR_FAILED_TO_INITIALIZE);
            LogMessage(LOG_ERROR_MESSAGE, errStr.Cstr());
         }
      }

      if (okayToAdd)
      {
         _transferList->AddItem(nextXfer);
         _transferList->ScrollTo(0, _transferList->CountItems()*TRANSFER_ITEM_HEIGHT);  // scroll to the bottom
      }
      else
      {
         RemoveHandler(nextXfer);
         delete nextXfer;  // he's toast
      }
   }
   DequeueTransferSessions();
}

void
ShareWindow ::
SendConnectBackRequestMessage(const char * sessionID, int port)
{
   _netClient->SendConnectBackRequestMessage(sessionID, port);
}


void
ShareWindow ::
DequeueTransferSessions()
{
   DequeueTransferSessions(true);
   DequeueTransferSessions(false);
}

uint32
ShareWindow ::
CountActiveSessions(bool upload, const char * optForSessionID) const
{
   uint32 numActive = 0;
   for (int i=_transferList->CountItems()-1; i>=0; i--)
   {
      ShareFileTransfer * next = (ShareFileTransfer *) _transferList->ItemAt(i);
      if ((next->IsUploadSession() == upload)&&(next->IsActive())&&((optForSessionID == NULL)||(strcmp(optForSessionID, next->GetRemoteSessionID()) == 0))) numActive++;
   }
   return numActive;
}

void
ShareWindow ::
DequeueTransferSessions(bool upload)
{
   int numToStart = (upload ? _maxSimultaneousUploadSessions : _maxSimultaneousDownloadSessions) - CountActiveSessions(upload, NULL);
   if (numToStart > 0)
   {
      uint32 numItems = _transferList->CountItems();
      for (uint32 j=0; j<numItems; j++)
      {
         ShareFileTransfer * next = (ShareFileTransfer *) _transferList->ItemAt(j);
         const char * nextSession = next->GetRemoteSessionID();
         if ((next->IsUploadSession() == upload)&&(next->IsWaitingOnLocal())&&
             ((upload == false)||(CountActiveSessions(true, nextSession[0] ? nextSession : NULL) < _maxSimultaneousUploadSessionsPerUser)))
         {
            next->BeginTransfer();
            if (--numToStart == 0) break;
         }
      }
   }
}

void ShareWindow :: OpenTrackerFolder(const BDirectory & dir)
{
   PortableString cmd("/boot/beos/system/Tracker ");
   BEntry ent;
   BPath p;
   if ((dir.GetEntry(&ent) == B_NO_ERROR)&&(ent.GetPath(&p) == B_NO_ERROR))
   {
      cmd += p.Path();
      system(cmd.Cstr());
   }
}


void
ShareWindow ::
LogHelp(const char * cmd, int tokenID, int descID)
{
   PortableString s("   /");
   s += cmd;
   s += " ";
   if (tokenID >= 0) s += str(tokenID);
   s += " - ";
   s += str(descID);
   LogMessage(LOG_INFORMATION_MESSAGE, s.Cstr());
}

void ShareWindow :: MessageReceived(BMessage * msg)
{
   switch(msg->what)
   {
      case SHAREWINDOW_COMMAND_CHECK_ALICE:
         if (_alice.Length() > 0)
         {
            _textEntry->SetText(_alice.Cstr());
            _alice = "";
            PostMessage(SHAREWINDOW_COMMAND_SEND_CHAT_TEXT);
         }
      break;

      case SHAREWINDOW_COMMAND_CHECK_FOR_MORIBUND_CONNECTIONS:
      {
         bigtime_t now = system_time();
         for (int i=_transferList->CountItems()-1; i>=0; i--)
         {
            ShareFileTransfer * next = (ShareFileTransfer *)_transferList->ItemAt(i);
            bigtime_t nextT = next->LastTransferTime();
            if (nextT > 0LL)
            {
               bigtime_t diff = now - nextT;
               if (diff > MORIBUND_TIMEOUT_SECONDS*1000000LL)
               {
                  printf("Connection %i is moribund, aborting!\n", i);
                  next->AbortSession(true);
               }
            } 
         }
      }
      break;

      case SHAREWINDOW_COMMAND_LAUNCH_TRANSFER_ITEM:
      {
         int32 nextIndex;
         for (int i=0; ((nextIndex = _transferList->CurrentSelection(i)) >= 0); i++) ((ShareFileTransfer *)_transferList->ItemAt(nextIndex))->LaunchCurrentItem();
      }
      break;

      case SHAREWINDOW_COMMAND_SELECT_LANGUAGE:
      {
         int32 l;
         if (msg->FindInt32("language", &l) == B_NO_ERROR)
         {
            _language = l;
            char temp[200];
            sprintf(temp, str(STR_LANGUAGE_SELECTED), GetLanguageName(_language));
            LogMessage(LOG_INFORMATION_MESSAGE, temp);
         }
      }
      break;
 
      case SHAREWINDOW_COMMAND_OPEN_SHARED_FOLDER:
         OpenTrackerFolder(_shareDir);
      break;

      case SHAREWINDOW_COMMAND_OPEN_DOWNLOADS_FOLDER:
         OpenTrackerFolder(_downloadsDir);
      break;

      case SHAREWINDOW_COMMAND_SELECT_USER:
      {
         PortableString strng = _fileNameQueryEntry->Text();
         int atIndex = strng.IndexOf('@');
         if (_fullUserQueries->IsMarked()) strng = "*";
                                      else if (atIndex >= 0) strng = strng.Substring(0, atIndex).Cstr();  // only the file part, pleez
         strng += '@';

         bool filesThere = false;
         int32 nextIndex;
         for (int i=0; ((nextIndex = _usersView->CurrentSelection(i)) >= 0); i++) 
         {
            RemoteUserItem * next = (RemoteUserItem *)_usersView->ItemAt(nextIndex);
            if (i > 0) strng += ',';
            strng += next->GetSessionID(); 
            if ((GetFirewalled() == false)||(next->GetFirewalled() == false)) filesThere = true;
         }
         if (filesThere)
         {
            _fileNameQueryEntry->SetText(strng.Cstr());
            bool wasEnabled = _queryEnabled;
            SetQueryEnabled(false, NULL);
            SetQueryEnabled(true, wasEnabled ? str(STR_CHANGED_QUERY_TO) : str(STR_STARTED_QUERY), false);
         }
      }
      break;

      case SHAREWINDOW_COMMAND_RESET_LAYOUT: 
         ResetLayout();
      break;

      case SHAREWINDOW_COMMAND_SET_ADVERTISED_BANDWIDTH:
      {
         const char * label;
         if ((msg->FindString("label", &label) == B_NO_ERROR)&&(msg->FindInt32("bps", (int32 *) &_uploadBandwidth) == B_NO_ERROR)) _netClient->SetUploadBandwidth(label, _uploadBandwidth);
      }
      break;

      case SHAREWINDOW_COMMAND_SET_UPLOAD_LIMIT:
         (void) msg->FindInt32("num", (int32 *) &_maxSimultaneousUploadSessions);
         DequeueTransferSessions();
      break;

      case SHAREWINDOW_COMMAND_SET_DOWNLOAD_LIMIT:
         (void) msg->FindInt32("num", (int32 *) &_maxSimultaneousDownloadSessions);
         DequeueTransferSessions();
      break;

      case SHAREWINDOW_COMMAND_SET_UPLOAD_PER_USER_LIMIT:
         (void) msg->FindInt32("num", (int32 *) &_maxSimultaneousUploadSessionsPerUser);
         DequeueTransferSessions();
      break;

      case SHAREWINDOW_COMMAND_CLEAR_CHAT_LOG:
         _chatText->Delete(0, _chatText->TextLength()-1);
         _clearChatItem->SetEnabled(false);
      break;

      case SHAREWINDOW_COMMAND_TOGGLE_FIREWALLED:
         if (!_isConnected)
         {
            _netClient->SetFirewalled(!_netClient->GetFirewalled());
            UpdateConnectStatus(false);
         }
      break;

      case SHAREWINDOW_COMMAND_TOGGLE_LOG_UPLOADS:
         _logUploads->SetMarked(!_logUploads->IsMarked());
      break;

      case SHAREWINDOW_COMMAND_TOGGLE_LOG_USER_EVENTS:
         _logUserEvents->SetMarked(!_logUserEvents->IsMarked());
      break;

      case SHAREWINDOW_COMMAND_TOGGLE_LOG_QUERY_CHANGES:
         _logQueryChanges->SetMarked(!_logQueryChanges->IsMarked());
      break;

      case SHAREWINDOW_COMMAND_TOGGLE_LOG_CHAT:
         _logChat->SetMarked(!_logChat->IsMarked());
      break;

      case SHAREWINDOW_COMMAND_TOGGLE_LOG_TIMESTAMPS:
         _logTimestamps->SetMarked(!_logTimestamps->IsMarked());
      break;

      case SHAREWINDOW_COMMAND_TOGGLE_FULL_USER_QUERIES:
         _fullUserQueries->SetMarked(!_fullUserQueries->IsMarked());
      break;

      case SHAREWINDOW_COMMAND_TOGGLE_BEEP_WHEN_MENTIONED:
         _beepWhenMentioned->SetMarked(!_beepWhenMentioned->IsMarked());
      break;

      case SHAREWINDOW_COMMAND_TOGGLE_AUTOCLEAR_COMPLETED_DOWNLOADS:
         _autoClearCompletedDownloads->SetMarked(!_autoClearCompletedDownloads->IsMarked());
      break;

      case SHAREWINDOW_COMMAND_NEW_PEER_SESSION:
      {
         int socket;
         if ((_shareServer)&&(_shareServer->GetNextIncomingConnection(socket) == B_NO_ERROR))
         {
            ShareFileTransfer * newSession = new ShareFileTransfer(_shareDir, _netClient->GetLocalSessionID());
            AddHandler(newSession);
            if (newSession->InitSocketUploadSession(socket, CountActiveSessions(true, NULL) >= _maxSimultaneousUploadSessions) == B_NO_ERROR) 
            {
               _transferList->AddItem(newSession);
            }
            else
            {
               LogMessage(LOG_ERROR_MESSAGE, str(STR_COULDNT_START_SHAREFILETRANSFER_SESSION));
               RemoveHandler(newSession);
               delete newSession;
            }
            UpdateDownloadButtonStatus();
            DequeueTransferSessions();
         }
      }
      break;

      case SHAREWINDOW_COMMAND_REMOVE_SESSION:
      {
         ShareFileTransfer * who;
         if (msg->FindPointer("who", (void **) &who) == B_NO_ERROR)
         {
//            bool isUploadSession = who->IsUploadSession();
            RemoveHandler(who);
            _transferList->RemoveItem(who);
            delete who;
            DequeueTransferSessions();
         }
      }
      break;

      case SHAREWINDOW_COMMAND_CLEAR_FINISHED_DOWNLOADS:
      {
         for (int i=_transferList->CountItems()-1; i>=0; i--)
         {
            ShareFileTransfer * next = (ShareFileTransfer *)_transferList->ItemAt(i);
            if (next->IsFinished())
            {
               RemoveHandler(next);
               _transferList->RemoveItem(i);
               delete next;  
            }
         }
         UpdateDownloadButtonStatus();
      }
      break;

      case SHAREWINDOW_COMMAND_RESULT_SELECTION_CHANGED:
         UpdateDownloadButtonStatus();
      break;

      case SHAREWINDOW_COMMAND_CANCEL_DOWNLOADS:
      {
         PortableQueue<ShareFileTransfer *> killList;
         int32 nextIndex;
         for (int i=0; ((nextIndex = _transferList->CurrentSelection(i)) >= 0); i++) 
         {
            ShareFileTransfer * next = (ShareFileTransfer *)_transferList->ItemAt(nextIndex);
            killList.AddTail(next);
         }
         for (int j=killList.CountItems()-1; j>=0; j--)
         {
            RemoveHandler(killList[j]);
            _transferList->RemoveItem(killList[j]);
            delete killList[j];
         }
         DequeueTransferSessions();
         UpdateDownloadButtonStatus();
      }
      break;

      case SHAREWINDOW_COMMAND_BEGIN_DOWNLOADS:
         RequestDownloads();
         _resultsView->DeselectAll();
         UpdateDownloadButtonStatus();
      break;
 
      case SHAREWINDOW_COMMAND_USER_CHANGED_NAME:
      {
         _netClient->SetLocalUserName(_userNameEntry->Text());
         if (_logUserEvents->IsMarked())
         {
            PortableString s(str(STR_YOUR_NAME_HAS_BEEN_CHANGED_TO));
            s += _netClient->GetLocalUserName();
            LogMessage(LOG_INFORMATION_MESSAGE, s.Cstr());
         }
         _textEntry->MakeFocus();
      }
      break;

      case SHAREWINDOW_COMMAND_DISCONNECT_FROM_SERVER:
         _netClient->DisconnectFromServer();
      break;

      case SHAREWINDOW_COMMAND_ABOUT:
      {
         char temp[200];
         sprintf(temp, "BeShare v1.63\n%s Jeremy Friesner\njaf@lcsaudio.com", str(STR_BY));
         (new BAlert("About BeShare", temp, "Hoopy"))->Go(NULL);
      }
      break;

      case SHAREWINDOW_COMMAND_SEND_CHAT_TEXT:
      {
         PortableString text = _textEntry->Text();
 
         // strip leading and trailing whitespace from text
         while(text.Length() > 0)
         {
            char c = text[text.Length()-1];
            if ((c == ' ')||(c == '\t')||(c=='\n')||(c=='\r')) text = text.Substring(0, text.Length()-1);
                                                          else break;
         }
         while((text.Length() > 0)&&((text[0] == ' ')||(text[0] == '\t'))) text = text.Substring(1);

         if ((text.Length() > 0)&&((_chatHistory.CountItems() == 0)||(_chatHistory.GetItemPointer(_chatHistory.CountItems()-1)->Equals(text) == false)))
         {
            _chatHistory.AddTail(text);
            if (_chatHistory.CountItems() > CHAT_HISTORY_LENGTH) _chatHistory.RemoveHead();
         }
         _chatHistoryPosition = -1;  // reset to bottom of list for next time

         if (text.StartsWith("/action ")) text = text.Substring(8).Prepend("/me ");
         if (text.StartsWith("/msg "))
         {
            bool sent = false;
            StringTokenizer tok(text.Cstr(), " ");
            (void)tok.GetNextToken();  // throw away the /msg
            const char * target = tok.GetNextToken();
            if (target)
            {
               const char * textMessage = strchr(target, '\0')+1;

               // See if it's a session ID
               RemoteUserItem * user;
               if (_users.Get(target, user)) 
               {
                  _netClient->SendChatMessage(target, textMessage);
                  LogMessage(LOG_LOCAL_USER_CHAT_MESSAGE, textMessage, target);
                  sent = true;
               }
               else
               {
                  // okay, probably a username then.  Find all users with that name.
                  PortableString tstr(target);
                  MakeRegexCaseInsensitive(tstr);
                  StringMatcher sm(tstr.Cstr());
                  PortableHashtableIterator<const char *, RemoteUserItem *> iter = _users.GetIterator();
                  while(iter.GetNextValue(user))
                  {
                     if (sm.Match(user->GetHandle())) 
                     {
                        _netClient->SendChatMessage(user->GetSessionID(), textMessage);
                        LogMessage(LOG_LOCAL_USER_CHAT_MESSAGE, textMessage, user->GetSessionID());
                        sent = true;
                     }
                  }
               }

               if (sent == false)  
               { 
                  PortableString temp(str(STR_UNKNOWN_USER));
                  temp += target;
                  temp += str(STR_MESSAGE);
                  temp += textMessage;
                  temp += str(STR_NOT_SENT);
                  LogMessage(LOG_ERROR_MESSAGE, temp.Cstr());
               }
            }
            else LogMessage(LOG_ERROR_MESSAGE, str(STR_NO_TARGET_USER_SPECIFIED_IN_MSG));
         }
         else if (text.StartsWith("/nick "))
         {
            _userNameEntry->SetText(text.Cstr()+6);
            PostMessage(SHAREWINDOW_COMMAND_USER_CHANGED_NAME);
         }
         else if (text.StartsWith("/clear"))
         {
            PostMessage(SHAREWINDOW_COMMAND_CLEAR_CHAT_LOG);
         }
         else if (text.StartsWith("/quit"))
         {
            PostMessage(B_QUIT_REQUESTED);
         }
         else if (text.StartsWith("/start"))
         {
            if (text.Length() > 7) _fileNameQueryEntry->SetText(text.Cstr()+7);
            PostMessage(SHAREWINDOW_COMMAND_CHANGE_FILE_NAME_QUERY);  // in case we have a query going
            PostMessage(SHAREWINDOW_COMMAND_ENABLE_QUERY);            // in case we don't
         }
         else if (text.StartsWith("/stop"))
         {
            PostMessage(SHAREWINDOW_COMMAND_DISABLE_QUERY);
         }
         else if (text.StartsWith("/disconnect"))
         {
            PostMessage(SHAREWINDOW_COMMAND_DISCONNECT_FROM_SERVER);
         }
         else if (text.StartsWith("/connect"))
         {
            if (text.Length() > 9) _serverEntry->SetText(text.Cstr()+9);
            PostMessage(SHAREWINDOW_COMMAND_RECONNECT_TO_SERVER);
         }
         else if (text.StartsWith("/help"))
         {
            LogMessage(LOG_INFORMATION_MESSAGE,                  str(STR_AVAILABLE_IRC_STYLE_COMMANDS));
            LogHelp("msg",        STR_TOKEN_NAME_OR_SESSION_ID_TEXT, STR_SEND_A_PRIVATE_MESSAGE);
            LogHelp("nick",       STR_TOKEN_NAME,                    STR_CHANGE_YOUR_USER_NAME);
            LogHelp("action",     STR_TOKEN_ACTION,                  STR_DO_SOMETHING);
            LogHelp("me",         STR_TOKEN_ACTION,                  STR_SYNONYM_FOR_ACTION);
            LogHelp("clear",      -1,                                STR_CLEAR_THE_CHAT_LOG);
            LogHelp("start",      STR_TOKEN_QUERY_STRING,            STR_START_A_NEW_QUERY);
            LogHelp("stop",       -1,                                STR_STOP_THE_CURRENT_QUERY);
            LogHelp("connect",    STR_TOKEN_SERVER_NAME,             STR_CONNECT_TO_A_SERVER);
            LogHelp("disconnect", -1,                                STR_DISCONNECT_FROM_THE_SERVER);
            LogHelp("quit",       -1,                                STR_QUIT_BESHARE);  
            LogHelp("help",       -1,                                STR_SHOW_THIS_HELP_TEXT);
         }
         else if (text.Length() > 0)
         {
            _netClient->SendChatMessage("*", text.Cstr());
            LogMessage(LOG_LOCAL_USER_CHAT_MESSAGE, text.Cstr());
         }
         _textEntry->SetText("");
      }
      break;

      case SHAREWINDOW_COMMAND_RECONNECT_TO_SERVER:
      {
         const char * server = _serverEntry->Text();
         if (server)
         {
            _connectedTo = server;  // save this for later, when we're connected
            StringTokenizer tok(server, " :");
            const char * host = tok.GetNextToken();
            if (host)
            {
               const char * portStr = tok.GetNextToken();
               int port = portStr ? atoi(portStr) : 0;
               if (port == 0) port = 2960;
               _netClient->ConnectToServer(host, port ? port : 2960);
            }
         }
         UpdateConnectStatus(true);
      }
      break;

      case SHAREWINDOW_COMMAND_CHANGE_FILE_NAME_QUERY:
      {
         const char * s;
         if (msg->FindString("query", &s) == B_NO_ERROR) _fileNameQueryEntry->SetText(s);
         if (_queryEnabled)
         {
            bool wasEnabled = _queryEnabled;
            SetQueryEnabled(false, NULL);
            SetQueryEnabled(true, wasEnabled ? str(STR_CHANGED_QUERY_TO) : str(STR_STARTED_QUERY));
         }
      }
      break;
 
      case SHAREWINDOW_COMMAND_USER_CHANGED_SERVER:  // user entered new server text
         UpdateConnectStatus(false);
      break;

      case SHAREWINDOW_COMMAND_USER_SELECTED_SERVER:  // user selected server from pop-up
      {
         const char * server;
         if (msg->FindString("server", &server) == B_NO_ERROR) 
         {
            bool reconnect = ((_isConnected == false)||(strcmp(server, _serverEntry->Text())));
            _serverEntry->SetText(server);
            if (reconnect) PostMessage(SHAREWINDOW_COMMAND_RECONNECT_TO_SERVER);
         }
      }
      break;

      case SHAREWINDOW_COMMAND_ENABLE_QUERY:
         _fileNameQueryEntry->MakeFocus(false);
         SetQueryEnabled(true, str(STR_STARTED_QUERY));
      break;

      case SHAREWINDOW_COMMAND_DISABLE_QUERY:
         SetQueryEnabled(false, str(STR_STOPPED_QUERY));
      break;

      case SHAREWINDOW_COMMAND_TOGGLE_COLUMN:
      {
         const char * attrName;
         if (msg->FindString("attrib", &attrName) == B_NO_ERROR)
         {
            BMenuItem * mi;
            if (_attribMenuItems.Get(attrName, mi))
            {
               mi->SetMarked(!mi->IsMarked());
               ShareColumn * sc;
               if (_columns.Get(attrName, sc))
               {
                  bool isVisible = (_resultsView->IndexOfColumn(sc) >= 0);
                  if ((isVisible)&&(mi->IsMarked() == false)) 
                  {
                     _activeAttribs.Remove(sc->GetAttributeName());
                     _resultsView->RemoveColumn(sc);
                  }
                  else if ((isVisible == false)&&(mi->IsMarked())) 
                  {
                     float w = DEFAULT_COLUMN_WIDTH;
                     if (_activeAttribs.Get(sc->GetAttributeName(), w)) sc->SetWidth(w);
                                                                   else _activeAttribs.Put(sc->GetAttributeName(), sc->Width());
                     _resultsView->AddColumn(sc);
                  }
               }
            }
         }
      }
      break;

      default:
         BWindow :: MessageReceived(msg);
      break;
   }
}

static PortableString _lastTalkedTo;

void
ShareWindow ::
LogMessage(LogMessageType type, const char * text, const char * optSessionID, const rgb_color * optTextColor, bool isPersonal)
{
   const char * preamble = "???"; 
   const rgb_color black = {0, 0, 0, 255};
   rgb_color color = black;  // default
   PortableString temp;
   int startRed = -1;
   int redLen = 0;
PortableString aliceResponse;
   
   if (optTextColor == NULL) optTextColor = &black;

   switch(type)
   {
      case LOG_INFORMATION_MESSAGE:
         preamble = str(STR_SYSTEM);
         color.blue = 128;
         _lastTalkedTo = "";
      break;

      case LOG_WARNING_MESSAGE:
         preamble = str(STR_WARNING);
         color.blue = color.green = 128;
         _lastTalkedTo = "";
      break;

      case LOG_ERROR_MESSAGE:
         preamble = str(STR_ERROR);
         color.red = 128;
         _lastTalkedTo = "";
      break;

      case LOG_LOCAL_USER_CHAT_MESSAGE:
         if (_logChat->IsMarked() == false) return;
         if (strncmp(text, "/me ", 4) == 0)
         {
            preamble = str(STR_ACTION);
            temp = _netClient->GetLocalUserName();
            temp += " ";
            temp += &text[4];
            text = temp.Cstr();
            color.red = color.blue = 128;
         }     
         else
         {
            temp = "(";
            temp += _netClient->GetLocalSessionID();
            temp += ") ";
            temp += _netClient->GetLocalUserName();
            if (optSessionID)
            {
               temp += "-> (";
               temp += optSessionID;
               temp += "/";
               temp += GetUserNameForSession(optSessionID);
               temp += ") ";
            }
            preamble = temp.Cstr();
            color.green = 128;
         }
      break;

      case LOG_REMOTE_USER_CHAT_MESSAGE:
         if (_logChat->IsMarked() == false) return;

         if (optSessionID)
         {
            if (strncmp(text, "/me ", 4) == 0)
            {
               preamble = str(STR_ACTION);
               temp = GetUserNameForSession(optSessionID);
               temp += " ";
               temp += &text[4];
               text = temp.Cstr();
               color.red = color.blue = 128;
            }
            else
            {
               temp = "(";
               temp += optSessionID;
               temp += ") ";
               temp += GetUserNameForSession(optSessionID);
               preamble = temp.Cstr();
               color.green = 1;  // if I don't put this, the text doesn't show up???

static char * answer = new char[4096];
static bool firstTime = true;
if (firstTime)
{
   (void) respond("sweety");
   firstTime = false;
}
answer = respond(text);
stripurl(answer);
aliceResponse = answer; 
            }
         }

         {
            // See if we were mentioned as part of the comment.
            // Only look for the first part of our username (before any
            // spaces or punctuation), as this is what people will often
            // shorten our name to.
            PortableString shortName(_netClient->GetLocalUserName());
            const char * orig = shortName.Cstr();
            const char * temp = orig;
            while(((*temp >= 'A')&&(*temp <= 'Z'))||((*temp >= 'a')&&(*temp <= 'z'))||((*temp >= '0')&&(*temp <= '9'))) temp++;
            if (temp > orig) shortName = shortName.Substring(0, temp-orig);
             
            PortableString iText(text);
            iText = iText.ToUpperCase();
            startRed = iText.IndexOf(shortName.ToUpperCase());
            if (startRed >= 0) 
            {
               redLen = shortName.Length();

               // See if we can't extend the red back out to the original name length....
               PortableString temp = _netClient->GetLocalUserName();
               temp = temp.ToUpperCase(); 
               const char * c1 = &iText.Cstr()[startRed+redLen];
               const char * c2 = &temp.Cstr()[redLen];
               while((c1)&&(c2)&&(*c1 == *c2))
               {
                  c1++;
                  c2++;
                  redLen++;
               }
            }
         }
         if ((isPersonal)||(startRed >= 0)||((optSessionID)&&(_lastTalkedTo == optSessionID)))
         {
            if (_beepWhenMentioned->IsMarked()) beep();

            if ((optSessionID)&&(aliceResponse.Length() > 0))
            {
               PortableString response;
               if (optSessionID)
               {
                  if (isPersonal)
                  {
                     response += "/msg ";
                     response += GetUserNameForSession(optSessionID);
                  }
                  else 
                  {
                     _lastTalkedTo = (strstr(text, "shut up") != NULL) ? "" : optSessionID;
                     if (drand48() > 0.75)
                     {
                        PortableString s(GetUserNameForSession(optSessionID));
                        int sp = s.IndexOf(' ');
                        if (sp < 0) sp = s.IndexOf('[');
                        if (sp < 0) sp = s.IndexOf('(');
                        if (sp < 0) sp = s.IndexOf(':');
                        if (sp < 0) sp = s.IndexOf('.');
                        if (sp > 0) s = s.Substring(0, sp);
                        response += s;
                        response += ": ";
                     }
                  }
                  response += " ";
               }
               response += aliceResponse; 
               _alice = response;
            }
         }
         else _lastTalkedTo = "";
      break;
   }

   text_run_array style;
   style.count = 1;
   style.runs[0].offset = 0;
   style.runs[0].font = *be_bold_font;
   style.runs[0].color = black;

   if (_logTimestamps->IsMarked())
   {
      time_t n = time(NULL);
      struct tm * now = localtime(&n);
      char buf[128];
      sprintf(buf, "[%i/%i %i:%02i] ", now->tm_mon+1, now->tm_mday, now->tm_hour, now->tm_min);
      _chatText->Insert(_chatText->TextLength(), buf, strlen(buf), &style);
   }

   style.runs[0].color = color;
   _chatText->Insert(_chatText->TextLength(), preamble, strlen(preamble), &style);

   // now back to black
   style.runs[0].color = *optTextColor;
   style.runs[0].font = *be_plain_font;

   _chatText->Insert(_chatText->TextLength(), ": ", 2, &style);
   if ((startRed >= 0)&&(redLen > 0))
   {
      _chatText->Insert(_chatText->TextLength(), text, startRed);
      const rgb_color red = {255, 128, 0, 255};
      style.runs[0].color = red;
      _chatText->Insert(_chatText->TextLength(), text+startRed, redLen, &style);
      style.runs[0].color = *optTextColor;
      _chatText->Insert(_chatText->TextLength(), text+startRed+redLen, strlen(text)-(startRed+redLen), &style);
   }
   else _chatText->Insert(_chatText->TextLength(), text, strlen(text));
   _chatText->Insert(_chatText->TextLength(), "\n", 2);

   BScrollBar * sb = _chatScrollView->ScrollBar(B_VERTICAL);
   if (sb) 
   {
      float min, max, smallStep, bigStep;
      sb->GetRange(&min, &max);
      sb->GetSteps(&smallStep, &bigStep);
      if (sb->Value() >= max-(smallStep*10)) sb->SetValue(max-smallStep);
   }

   _clearChatItem->SetEnabled(true);
}

 
const char *
ShareWindow ::
GetUserNameForSession(const char * sessionID) const
{
   RemoteUserItem * user;
   return (_users.Get(sessionID, user)) ? user->GetHandle() : str(STR_UNKNOWN);
}

void
ShareWindow ::
SetQueryEnabled(bool e, const char * logTitle, bool putInQueryMenu)
{
   if (e != _queryEnabled)
   {
      _queryEnabled = e;
      if (_queryEnabled) 
      {
         // If there is an '@' sign, split the string into separate filename and username queries
         PortableString fileExp(_fileNameQueryEntry->Text());
         PortableString userExp;   // default == empty == "*"

         if (putInQueryMenu)
         {
            // Put the query into the query list, or move it to the top
            int idx = -1;
            for (int i=_queryMenu->CountItems()-1; i>=0; i--)
            {
               if (strcmp(_queryMenu->ItemAt(i)->Label(), fileExp.Cstr()) == 0)
               {
                  // switch this item with the first item in the menu
                  if (i > 0)
                  {
                     BMenuItem * bottom = _queryMenu->RemoveItem(i);
                     BMenuItem * top    = _queryMenu->RemoveItem((int32)0);
                     _queryMenu->AddItem(bottom, 0);
                     _queryMenu->AddItem(top, i);
                  }
                  idx = i;
                  break;
               }
            }
            if (idx == -1)
            {
               // add item to end of menu
               BMessage * msg = new BMessage(SHAREWINDOW_COMMAND_CHANGE_FILE_NAME_QUERY);
               msg->AddString("query", fileExp.Cstr());
               _queryMenu->AddItem(new BMenuItem(fileExp.Cstr(), msg), 0);
            }

            // Keep it reasonable
            while(_queryMenu->CountItems() > 20) delete _queryMenu->RemoveItem(20);
         }

         int32 atIndex = fileExp.IndexOf('@');
         if (atIndex >= 0)
         {
            if ((uint32) atIndex < fileExp.Length()) 
            {
               userExp = fileExp.Substring(atIndex+1);  // in case they entered a session ID
             
               // Since the user probably entered a user name instead of a
               // session ID, let's go down our user list and find any matching
               // names, and add their IDs to the search.
               if (HasRegexTokens(userExp.Cstr()) == false) userExp = userExp.Prepend("*").Append("*"); 
               MakeRegexCaseInsensitive(userExp);
               StringMatcher match(userExp.Cstr());
               PortableHashtableIterator<const char *, RemoteUserItem *> iter = _users.GetIterator();
               RemoteUserItem * next;
               while(iter.GetNextValue(next)) 
               {
                  if (match.Match(next->GetHandle())) 
                  { 
                     userExp += ",";
                     userExp += next->GetSessionID();
                  }
               }
            }

            if (atIndex > 0) fileExp = fileExp.Substring(0, atIndex);
                        else fileExp = "";  // i.e. "*"
         }

         // If there are regexp chars in the filename, use it verbatim;
         // otherwise add *'s around the edges to make it a substring search.
         if (HasRegexTokens(fileExp.Cstr()) == false) fileExp = fileExp.Prepend("*").Append("*");

         // Don't allow slashes in the queries as that would screw things up
         fileExp.Replace('/', '?');
         userExp.Replace('/', '?');

         MakeRegexCaseInsensitive(fileExp);

         ClearResults();
         _netClient->StartQuery(userExp.Length() > 0 ? userExp.Cstr() : "*", fileExp.Cstr());

         if ((logTitle)&&(_logQueryChanges->IsMarked()))
         {
            PortableString str(logTitle);
            str += ": [";
            str += _fileNameQueryEntry->Text(); 
            str += "]";
            LogMessage(LOG_INFORMATION_MESSAGE, str.Cstr());
         }
      }
      else 
      {
         _netClient->StopQuery();
         if ((logTitle)&&(_logQueryChanges->IsMarked())) LogMessage(LOG_INFORMATION_MESSAGE, logTitle);
      }

      UpdateQueryEnabledStatus();
   }
}

void
ShareWindow ::
UpdateQueryEnabledStatus()
{
   _enableQueryButton->SetEnabled((_isConnected)&(!_queryEnabled));
   _disableQueryButton->SetEnabled((_isConnected)&(_queryEnabled));
}

void
ShareWindow ::
UpdateConnectStatus(bool titleToo)
{
   const char * sname = _serverEntry->Text();

   bool c = ((_isConnected)||(_isConnecting));
   _connectMenuItem->SetEnabled(!c);
   if (sname[0] == '\0') sname = "???";
   _disconnectMenuItem->SetEnabled(c);

   _firewalled->SetEnabled(!_isConnected);
   _firewalled->SetMarked(_netClient->GetFirewalled());

   if (titleToo) UpdateTitleBar();

   char buf[200];
   strcpy(buf, str(STR_CONNECT_TO));
   strncat(buf, sname, sizeof(buf));
   buf[sizeof(buf)-1] = '\0';
   _connectMenuItem->SetLabel(buf);

   UpdateDownloadButtonStatus();
}

void
ShareWindow ::
UpdateTitleBar()
{
   char buf[200];
   strcpy(buf, "BeShare");
   if ((_isConnecting)&&(_isConnected == false))
   {
      strncat(buf, str(STR_CONNECTING_TO), sizeof(buf)); 
      strncat(buf, _connectedTo.Cstr(), sizeof(buf));
   }
   else 
   {
      uint32 numFiles = _resultsView->CountItems();
      if (numFiles > 0)
      {
         char buf2[100];
         sprintf(buf2, str(STR_PERCENTLU_RESULTS_SHOWN), numFiles);
         strncat(buf, buf2, sizeof(buf));
         char * nul = strchr(buf, '\0');
         GetByteSizeString(_bytesShown, nul);
         strncat(buf, ")", sizeof(buf));
      }
      else if (_isConnected)
      {
         strncat(buf, str(STR_CONNECTED_TO), sizeof(buf)); 
         strncat(buf, _connectedTo.Cstr(), sizeof(buf));
      }
   }

   buf[sizeof(buf)-1] = '\0';
   SetTitle(buf);
}

void
ShareWindow ::
UpdateDownloadButtonStatus()
{
   _requestDownloadsButton->SetEnabled((_isConnected)&&(_resultsView->CurrentSelection() >= 0));

   bool deadTransfersPresent = false;
   for (int i=_transferList->CountItems()-1; i>=0; i--)
   {
      ShareFileTransfer * next = (ShareFileTransfer *)_transferList->ItemAt(i);
      if (next->IsFinished())
      {
         deadTransfersPresent = true;
         break;
      }
   }
   _clearFinishedDownloadsButton->SetEnabled(deadTransfersPresent);
   _cancelTransfersButton->SetEnabled(_transferList->CurrentSelection() >= 0);
}

bool
ShareWindow ::
QuitRequested()
{
   be_app->PostMessage(B_QUIT_REQUESTED);
   return true;
}


// The methods below are called by our ShareNetClient at the appropriate times.
void 
ShareWindow ::
SetConnectStatus(bool isConnecting, bool isConnected)
{
   if ((!_isConnected)&&(isConnected)) 
   {
      LogMessage(LOG_INFORMATION_MESSAGE, str(STR_CONNECTION_ESTABLISHED));

      // If the current server isn't in our server menu, add it now
      bool found = false;
      const char * serverName = _connectedTo.Cstr();
      for (int i=_serverMenu->CountItems()-1; i>=0; i--)
      {
         if (strcmp(_serverMenu->ItemAt(i)->Label(), serverName) == 0)
         {
            found = true;
            break;
         }
      }
      if (found == false) AddServerItem(serverName);
   }
   else if ((!_isConnecting)&&(isConnecting)) LogMessage(LOG_INFORMATION_MESSAGE, str(STR_CONNECTING_TO_SERVER_DOTDOTDOT));
   else if ((isConnecting == false)&&(isConnected == false))
   {
           if (_isConnected) LogMessage(LOG_ERROR_MESSAGE, str(STR_YOU_ARE_NO_LONGER_CONNECTED_TO_THE_MUSCLE_SERVER));
      else if (_isConnecting) LogMessage(LOG_ERROR_MESSAGE, str(STR_CONNECTION_TO_SERVER_FAILED));
   }

   _isConnecting = isConnecting;
   _isConnected  = isConnected;

   // If we're not connected anymore, make sure the display is clear
   if (_isConnected == false) 
   {
      ClearUsers();
      SetQueryEnabled(false, NULL);
   }

   UpdateConnectStatus(true);
   UpdateQueryEnabledStatus();
}


void 
ShareWindow ::
PutUser(const char * sessionID, const char * userName, const char * hostName, int port)
{
   RemoteUserItem * user;
   if (_users.Get(sessionID, user) == false)
   {
      user = new RemoteUserItem(this, sessionID);
      _users.Put(user->GetSessionID(), user);
      _usersView->AddItem(user);
      _usersView->SortItems();
   }
   if (userName) user->SetHandle(userName);
   if (hostName) user->SetHostName(hostName);
   if (port >= 0) user->SetPort(port);
}


void
ShareWindow ::
SetUserBandwidth(const char * sessionID, const char * label, uint32 bps)
{
   PutUser(sessionID, NULL, NULL, -1);  // make sure the RemoteUserItem is present!
   RemoteUserItem * user;
   if (_users.Get(sessionID, user)) user->SetBandwidth(label, bps);
}

void
ShareWindow ::
SetUserIsFirewalled(const char * sessionID, bool fw)
{
   PutUser(sessionID, NULL, NULL, -1);  // make sure the RemoteUserItem is present!
   RemoteUserItem * user;
   if (_users.Get(sessionID, user)) user->SetFirewalled(fw);
}

void
ShareWindow ::
SetUserFileCount(const char * sessionID, int32 fc)
{
   PutUser(sessionID, NULL, NULL, -1);  // make sure the RemoteUserItem is present!
   RemoteUserItem * user;
   if (_users.Get(sessionID, user)) user->SetNumSharedFiles(fc);
}

void 
ShareWindow ::
RemoveUser(const char * sessionID)
{
   RemoteUserItem * user;
   if (_users.Remove(sessionID, user)) 
   {
      _usersView->RemoveItem(user);
      delete user;
   } 
}

void 
ShareWindow ::
PutResult(const char * sessionID, const char * fileName, bool isFirewalled, PortableMessageRef fileInfo)
{
   PutUser(sessionID, NULL, NULL, -1);  // make sure the RemoteUserItem is present!
   RemoteUserItem * user;
   if (_users.Get(sessionID, user)) 
   {
      user->SetFirewalled(isFirewalled);
      user->PutFile(fileName, fileInfo);
   }
}

void 
ShareWindow ::
RemoveResult(const char * sessionID, const char * fileName)
{
   RemoteUserItem * user;
   if (_users.Get(sessionID, user)) user->RemoveFile(fileName);
}

void 
ShareWindow ::
ClearResults()
{
   _resultsView->MakeEmpty();  // for efficiency
   PortableHashtableIterator<const char *, RemoteUserItem *> iter = _users.GetIterator();
   RemoteUserItem * next;
   while(iter.GetNextValue(next)) next->ClearFiles();
   _bytesShown = 0LL;
   UpdateTitleBar();
}

void
ShareWindow ::
FileTransferConnected(ShareFileTransfer * who)
{
   RefreshTransferItem(who);
   UpdateDownloadButtonStatus();
}

void
ShareWindow ::
BeginBatchFileResultUpdate()
{
   // Do everything we can to make the following results-list updates
   // happen quickly, as there may be a lot of them...
   DisableUpdates();
}

void
ShareWindow ::
EndBatchFileResultUpdate()
{
   if (_tempAddList.CountItems() > 0)
   {
      _resultsView->AddList(&_tempAddList);
      _resultsView->SortItems();
      _tempAddList.MakeEmpty();
   }
   EnableUpdates();
   UpdateTitleBar();
}

void
ShareWindow ::
FileTransferDisconnected(ShareFileTransfer * who)
{
   // Uploads we just immediately get rid of, downloads
   // stick around till the user removes them.
   if ((who->IsUploadSession())||((_autoClearCompletedDownloads->IsMarked())&&(who->ErrorOccurred() == false)))
   {
      // Can't just delete it here because it is calling
      // me!  So I'll send myself a message to do it later
      BMessage msg(SHAREWINDOW_COMMAND_REMOVE_SESSION);
      msg.AddPointer("who", who);
      PostMessage(&msg);
   }
   else 
   {
      DequeueTransferSessions();
      RefreshTransferItem(who);
   }

   UpdateDownloadButtonStatus();
}


const char * 
ShareWindow ::
GetFileCellText(const RemoteFileItem * item, int32 columnIndex) const
{
   return ((ShareColumn *)_resultsView->ColumnAt(columnIndex))->GetFileCellText(item);
}

const BBitmap *
ShareWindow ::
GetBitmap(const RemoteFileItem * item, int32 /*columnIndex*/) const
{
   const BBitmap * bmp = NULL;

   // For now, the only bitmap supported is the MIME type icon
   const char * mimeString;
   if (item->GetAttributes().FindString("beshare:Kind", &mimeString) == B_NO_ERROR)
   {
      ShareMIMEInfo * mi;
      if (_mimeInfos.Get(mimeString, mi)) bmp = mi->GetIcon();
   }
   return bmp ? bmp : &_defaultBitmap;
}

const BBitmap *
ShareWindow ::
GetBitmap(const char * mimeString)
{
   ShareMIMEInfo * mi = mimeString ? CacheMIMETypeInfo(mimeString) : NULL;
   const BBitmap * bmp = mi ? mi->GetIcon() : NULL;
   return bmp ? bmp : &_defaultBitmap;
}

ShareMIMEInfo *
ShareWindow ::
CacheMIMETypeInfo(const char * mimeString)
{
   ShareMIMEInfo * ret;
   if (_mimeInfos.Get(mimeString, ret) == false)
   {
      const char * label = mimeString;
      char buf[B_MIME_TYPE_LENGTH];
      BMimeType mt(mimeString);
      if ((mt.InitCheck()==B_NO_ERROR)&&(mt.GetShortDescription(buf) == B_NO_ERROR)) label = buf;

      ShareMIMEInfo * newInfo = new ShareMIMEInfo(label, mimeString);
      _mimeInfos.Put(newInfo->GetMIMEString(), newInfo);
      _emptyMimeInfos.Put(newInfo, true);   // because it's not in the menu yet (it goes there when we have something to put in it)
      ret = newInfo;
   }
   return ret;
}

void
ShareWindow :: 
AddFileItem(RemoteFileItem * item)
{
   _tempAddList.AddItem(item);

   const PortableMessage & attrs = item->GetAttributes();
   ShareMIMEInfo * optMimeInfo = NULL;
   const char * mimeString = NULL;
   if (attrs.FindString("beshare:Kind", &mimeString) == B_NO_ERROR) optMimeInfo = CacheMIMETypeInfo(mimeString);

   int64 s;
   if (attrs.FindInt64("beshare:File Size", &s) == B_NO_ERROR) _bytesShown += s;

   PortableMessageFieldNameIterator iter = attrs.GetFieldNameIterator();
   const char * next;
   while((next = iter.GetNextFieldName()) != NULL) CreateColumn(optMimeInfo, next, true, 0);
}



void
ShareWindow ::
CreateColumn(ShareMIMEInfo * optMimeInfo, const char * attrName, bool remote, char key)
{
   if (_columns.ContainsKey(attrName) == false) 
   {
      const char * label = attrName;
      int type = ShareColumn::ATTR_MISC;

      if ((remote == false)&&(attrName[0] == SPECIAL_COLUMN_CHAR))
      {
         type = attrName[1]-'0';
         label += 2;
      }
      else
      {
         if (strncmp(attrName, "beshare:", 8) == 0)
         {
            // A BeShare var; handle these a bit differently  
            label += 8;
                 if (strcmp(label, "Kind")     ==0) label = str(STR_KIND);       // hacked-in language support :^P
            else if (strcmp(label, "File Size")==0) label = str(STR_FILE_SIZE);
            optMimeInfo = NULL;   // these vars are not type specific
         }
         else
         {
            // It's a genuine attribute; try to find a better name for it
            if (optMimeInfo) 
            {
               const char * desc = optMimeInfo->GetAttributeDescription(attrName);
               if (desc) label = desc;
            }
         }
      }

      ShareColumn * column = new ShareColumn(type, attrName, label, 80.0f);
      _columns.Put(column->GetAttributeName(), column);

      BMessage * msg = new BMessage(SHAREWINDOW_COMMAND_TOGGLE_COLUMN);
      msg->AddString("attrib", column->GetAttributeName());
      BMenuItem * mi = new BMenuItem(column->GetLabel(), msg, key);

      if (optMimeInfo) 
      {
         if (_emptyMimeInfos.ContainsKey(optMimeInfo))
         {
            _attribMenu->AddItem(optMimeInfo);    // only add it to our menu when it finally gets something to hold
            _emptyMimeInfos.Remove(optMimeInfo);  // now the BMenu is responsible for deleting it
         }
         optMimeInfo->AddItem(mi);
      }
      else _attribMenu->AddItem(mi);

      _attribMenuItems.Put(column->GetAttributeName(), mi);
      if (_activeAttribs.ContainsKey(attrName)) PostMessage(msg);
   }
}


void
ShareWindow :: RemoveFileItem(RemoteFileItem * item)
{
   int64 s;
   if (item->GetAttributes().FindInt64("beShare:File Size", &s) == B_NO_ERROR) _bytesShown -= s;

   _resultsView->RemoveItem(item);
}

void
ShareWindow ::
RefreshTransferItem(ShareFileTransfer * item)
{
   _transferList->InvalidateItem(_transferList->IndexOf(item));
}


void
ShareWindow ::
RefreshFileItem(RemoteFileItem * item)
{
   _resultsView->InvalidateItem(_resultsView->IndexOf(item));
}

void
ShareWindow ::
RefreshUserItem(RemoteUserItem * item)
{
   _usersView->InvalidateItem(_usersView->IndexOf(item));
   _usersView->SortItems();
}


int
ShareWindow ::
CompareFunc(const CLVListItem* item1, const CLVListItem* item2, int32 sortKey)
{
   return ((RemoteFileItem *)item1)->Compare(((RemoteFileItem *)item2), sortKey);  
}

int ShareWindow ::
UserCompareFunc(const CLVListItem * i1, const CLVListItem * i2, int32 sortKey)
{
   const RemoteUserItem * u1 = (const RemoteUserItem *) i1;
   const RemoteUserItem * u2 = (const RemoteUserItem *) i2;
   switch(sortKey)
   {
      case 0:  return strcmp(u1->GetHandle(), u2->GetHandle());
      case 1:  return strcmp(u1->GetSessionID(), u2->GetSessionID());
      case 2:  return (u1->GetNumSharedFiles() > u2->GetNumSharedFiles()) ? -1 : ((u1->GetNumSharedFiles() == u2->GetNumSharedFiles()) ? 0 : 1);
      case 3:  return (u1->GetBandwidth() > u2->GetBandwidth()) ? -1 : ((u1->GetBandwidth() == u2->GetBandwidth()) ? 0 : 1);
      default: return 0;
   } 
}


int 
ShareWindow ::
Compare(const RemoteFileItem * rf1, const RemoteFileItem * rf2, int32 sortKey) const
{
   return ((ShareColumn *)_resultsView->ColumnAt(sortKey))->Compare(rf1, rf2);
}


bool
ShareWindow ::
LogUploads() const
{
   return _logUploads->IsMarked();
}


bool
ShareWindow ::
LogUserEvents() const
{
   return _logUserEvents->IsMarked();
}

void 
ShareWindow ::
ResetLayout()
{
   _resultsTransferSplit->SetAlignment(B_VERTICAL);
   _chatUsersSplit->SetAlignment(B_VERTICAL);
   _mainSplit->SetAlignment(B_HORIZONTAL);

}

filter_result 
ShareWindow ::
Filter(BMessage * msg, BHandler ** target)
{
   if (msg->what == B_KEY_DOWN)
   {
      int8 c;
      int32 mod;
      if ((msg->FindInt8("byte", &c) == B_NO_ERROR)&&
          (msg->FindInt32("modifiers", &mod) == B_NO_ERROR)&&((mod & B_LEFT_COMMAND_KEY) == 0))
      {
         if (*target == _textEntry->TextView())
         {
            if (c == B_ENTER) PostMessage(SHAREWINDOW_COMMAND_SEND_CHAT_TEXT);
            else
            {
               int lineDiff = 0;
                    if (c == B_UP_ARROW)   lineDiff = (mod & B_SHIFT_KEY) ? -CHAT_HISTORY_LENGTH : -1;
               else if (c == B_DOWN_ARROW) lineDiff = (mod & B_SHIFT_KEY) ? CHAT_HISTORY_LENGTH  :  1;

               if ((lineDiff)&&(_chatHistory.CountItems() > 0))
               {
                  int numItems = _chatHistory.CountItems();
                  if (_chatHistoryPosition == -1)
                  {
                     if (lineDiff < 0) _chatHistoryPosition = numItems;
                                  else lineDiff = 0;  // at bottom + down arrow = no effect
                  }

                  if (lineDiff)
                  {
                     _chatHistoryPosition += lineDiff;
                     if (_chatHistoryPosition < 0) _chatHistoryPosition = 0;
                     if (_chatHistoryPosition >= numItems) 
                     {
                        _chatHistoryPosition = -1;
                        _textEntry->SetText("");
                     }
                     else
                     {
                        const PortableString * str = _chatHistory.GetItemPointer(_chatHistoryPosition);
                        _textEntry->SetText(str->Cstr());
                        _textEntry->TextView()->Select(str->Length(), str->Length());
                     }
                     return B_SKIP_MESSAGE;
                  }
               }
            }
         }
         else
         {
            if (c >= ' ')
            {
               _textEntry->MakeFocus();

               PortableString str(_textEntry->Text());
               str += c;
               _textEntry->SetText(str.Cstr());
               _textEntry->TextView()->Select(str.Length(), str.Length());
               return B_SKIP_MESSAGE;
            }
            else if ((c == B_ENTER)&&(*target == _resultsView)) PostMessage(SHAREWINDOW_COMMAND_BEGIN_DOWNLOADS);
         }
      }
   }
   else if (msg->what == B_MOUSE_DOWN)
   {
           if ((*target == _resultsView)&&(_resultsView->IsFocus() == false)) _resultsView->MakeFocus();
      else if ((*target == _usersView)&&(_usersView->IsFocus() == false)) _usersView->MakeFocus();
      else if ((*target == _chatText)&&(_chatText->IsFocus() == false)) _chatText->MakeFocus();
   }
   return B_DISPATCH_MESSAGE;
}

bool
ShareWindow ::
GetFirewalled() const
{
   return _netClient->GetFirewalled();
}
